# pylint: disable=no-member, arguments-differ
# pylint: disable=unsubscriptable-object
"""API Entities Module

This module provides a comprehensive set of Pydantic models that represent
multiple entities obtained from the API . 

Classes:
    - Dataset: Represents a structure for datasets present in the models.
    - Model: Represents a structure for models fetched from the API.
    - Project: Represents a structure for projects retrieved from the API.
"""

from abc import ABC, abstractmethod
from typing import Literal, Optional, Type, Union

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, model_validator

from dhl_sdk._input_processing import (
    CultivationHistoricalPreprocessor,
    CultivationPropagationPreprocessor,
    Preprocessor,
    SpectraData,
    SpectraPreprocessor,
    format_predictions,
)
from dhl_sdk._constants import (
    DATASETS_URL,
    MODELS_URL,
    PREDICT_URL,
    PROJECTS_URL,
    TEMPLATES_URL,
)
from dhl_sdk._utils import (
    PredictionRequestConfig,
    Predictions,
    PredictionResponse,
    get_id_list,
)
from dhl_sdk.crud import Client, CRUDClient, DataBaseClient, Result
from dhl_sdk.db_entities import Experiment, Variable
from dhl_sdk.exceptions import (
    InvalidInputsException,
    ModelPredictionException,
    PredictionRequestException,
)


# pylint: disable=not-an-iterable
class Dataset(BaseModel):
    """Model Dataset"""

    id: str = Field(alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")
    variables: list[Variable] = Field(alias="variables")
    experiments: list[Experiment] = Field(default=[], alias="experiments")
    _client: Client = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    def get_data(self, client: DataBaseClient) -> list[dict]:
        """Get the data from the experiments

        Returns:
        --------
            list[dict]: List with the experiment data organized by:
                key: variable code
                value: dictionary with 2 keys:
                    - "values": list with the values
                    - "timestamps": list with the timestamps for each value
        """

        run_data = []
        for experiment in self.experiments:
            data = experiment.get_data(client=client)
            run_data.append(data)

        return run_data

    @model_validator(mode="before")
    @classmethod
    def _unpack_reference_entities(cls, data) -> dict:
        def unpack_entity(source: str, entity):
            unpacked = []

            for i, info in enumerate(data[source]):
                try:
                    entity_id = info["id"]
                except KeyError as err:
                    raise KeyError(
                        f"The {source} index {i} does not contain an id"
                    ) from err

                var = entity.requests(data["client"]).get(entity_id)
                unpacked.append(var)

            data[source] = unpacked

        unpack_entity("variables", Variable)
        unpack_entity("experiments", Experiment)

        return data

    @staticmethod
    def requests(client: Client) -> CRUDClient["Dataset"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["Dataset"](client, DATASETS_URL, Dataset)


class SpectraDataset(Dataset):
    """Pydantic Model Dataset for Spectra"""

    _client: Client = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    def get_spectra_index(self) -> int:
        """Get the index of the spectra variable"""
        for index, variable in enumerate(self.variables):
            if variable.variant == "spectrum":
                return index
        raise ValueError("No spectrum variable found in dataset")

    def get_spectra_code(self) -> str:
        """Get variable code of spectra variable"""
        for variable in self.variables:
            if variable.variant == "spectrum":
                return variable.code
        raise ValueError("No spectrum variable found in dataset")

    @staticmethod
    def requests(client: Client) -> CRUDClient["SpectraDataset"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["SpectraDataset"](client, DATASETS_URL, SpectraDataset)


class Model(BaseModel, ABC):
    """Pydantic BaseModel for predictive models from the API"""

    id: str = Field(alias="id")
    name: str = Field(alias="name")
    status: str = Field(alias="status")
    project_id: str = Field(alias="projectId")
    dataset: Dataset = Field(alias="dataset")
    config: dict = Field(alias="config")
    _client: Client = PrivateAttr()

    @property
    def success(self) -> bool:
        """Get the success status of the model"""
        return self.status == "success"

    @staticmethod
    @abstractmethod
    def requests(client: Client) -> CRUDClient["Model"]:
        """Resquests abstract method for Model Types"""

    def get_predictions(self, preprocessor: Preprocessor) -> dict:
        """Get the predictions for the model using selected strategy"""

        if preprocessor.validate():
            json_data = preprocessor.format()
        else:
            raise InvalidInputsException(
                "The provided inputs failed the validation step"
            )

        predictions = []
        for prediction_data in json_data:
            try:
                response = self._client.post(PREDICT_URL, prediction_data)
                response.raise_for_status()

                # in case of an error in the response (not HTTP)
                if "error" in response.json():
                    raise PredictionRequestException(response.json()["error"])

            except Exception as ex:
                raise ex

            predictions.append(PredictionResponse(**response.json()))

        return format_predictions(predictions, model=self)

    @property
    def model_variables(self) -> dict:
        """List of the variables used in the model"""

        model_variables = []
        groups: dict = self.config["groups"]

        for variable in self.dataset.variables:
            variable_id = variable.id
            for group, variables in groups.items():
                if variable_id in variables:
                    model_variables.append(variable)
                    if group == "Flows":
                        feed_concentrations = [
                            ref.concentration_id
                            for ref in variable.variant_details.references
                            if ref.concentration_id
                        ]
                        if feed_concentrations:
                            groups.update({variable_id: feed_concentrations})
                    break

        return model_variables

    def get_model_variables_codes(self) -> list[str]:
        """Get the codes of the variables used in the model"""

        return [variable.code for variable in self.model_variables]

    @abstractmethod
    def predict(self, **kwargs) -> Predictions:
        """Prediction for Model"""


class PredictionConfig(BaseModel):
    """Configuration class for prediction method. This configuration parameters are
    passed to the model in DHL.

    Parameters:
    -----------

    model_confidence: float, optional
        This parameter determines the range within which the predictions of the model are
        expected to fall, with a specified level of certainty, i.e, setting it to 80%
        corresponds to capturing the range between the 10th and 90th percentiles
        of the model's output. Must be a value between 1 and 99, by default 80
    starting_index: int, optional
        The index of the timestamp after which the prediction will commence
    """

    model_config = ConfigDict(protected_namespaces=())

    model_confidence: float = Field(default=80.0, ge=1.0, le=99.0)
    starting_index: int = Field(default=0, ge=0)


class SpectraModel(Model):
    """Pydantic Model for Spectra Prediction Model from the API"""

    dataset: SpectraDataset = Field(alias="dataset")
    _spectra_size: Optional[int] = PrivateAttr(default=None)

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    @model_validator(mode="before")
    @classmethod
    def _validate_model_data(cls, data) -> dict:
        data["dataset"] = SpectraDataset(**data["dataset"], client=data["client"])
        return data

    def predict(
        self,
        spectra: SpectraData,
        inputs: Optional[dict] = None,
    ) -> Predictions:
        """
        Predicts the output of a given model for a given set of spectra.

        Parameters:
        -----------
        spectra : list[list[float]] or np.ndarray
            A 2D array representing spectra for prediction, where:
                - The first dimension corresponds to the spectra index.
                - The second dimension contains wavelength index.
        inputs : dict, optional
            Additional inputs to be used for prediction. The keys must be the Codes of the
            input variables, and the values must be lists of the same length as the number
            of spectrum, by default None.

        Returns:
        --------
        Dictionary with predictions where:
            key: variable code
            value: list with predictions for each spectrum

        Example:
        --------
        >>> spectra = [[1.0, 2.0, 3.0, 3.0, 4.0, 3.0], [2.0, 3.0, 4.0, 4.0, 5.0, 4.0]]
        >>> inputs = {"input1": [42, 33]}
        >>> result = model.predict(spectra, inputs)
        >>> print(result)
        {'output1': [prediction1, prediction2], 'output2': [prediction1, prediction2]}

        """

        if not self.success:
            raise ModelPredictionException(
                f"{self.name} is not ready for prediction. The current status is {self.status}"
            )

        spectra_processing_strategy = SpectraPreprocessor(
            spectra=spectra, inputs=inputs, model=self
        )

        predictions = super().get_predictions(spectra_processing_strategy)

        spectra_code = self.dataset.get_spectra_code()
        if spectra_code in predictions:
            predictions.pop(spectra_code)

        return predictions

    @property
    def inputs(self) -> list[str]:
        """Get the inputs from the model's config"""
        return self.config["groups"]["Inputs"]

    @property
    def outputs(self) -> list[str]:
        """Get the outputs from the model's config"""
        return self.config["groups"]["Outputs"]

    @property
    def spectra_size(self) -> int:
        """Get the size of the spectra"""
        if self._spectra_size is None:
            self._spectra_size = self._get_spectra_size()
        return self._spectra_size

    def _get_spectra_size(self) -> int:
        """Get the size of the spectra from variable information in the API"""
        spectrum = self.dataset.variables[self.dataset.get_spectra_index()]
        return spectrum.size

    @staticmethod
    def requests(client: Client) -> CRUDClient["SpectraModel"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["SpectraModel"](client, MODELS_URL, SpectraModel)


class CultivationModel(Model, ABC):
    """Abstract Pydantic Model for Cultivation Prediction Model from the API"""

    @model_validator(mode="before")
    @classmethod
    def _validate_model_data(cls, data) -> dict:
        data["dataset"] = Dataset(**data["dataset"], client=data["client"])
        return data

    @abstractmethod
    def predict(
        self,
        timestamps: list,
        inputs: dict,
        timestamps_unit: str = "s",
        config: PredictionConfig = PredictionConfig(),
    ) -> dict:
        """Prediction for CultivationModel"""

    @staticmethod
    @abstractmethod
    def requests(client: Client) -> CRUDClient["CultivationModel"]:
        """CRUDClient for CultivationModel"""


class CultivationPropagationModel(CultivationModel):
    """Pydantic Model for Propagation Model for Cultivation from the API"""

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    def predict(
        self,
        timestamps: list,
        inputs: dict,
        timestamps_unit: str = "s",
        config: PredictionConfig = PredictionConfig(),
    ) -> dict:
        """
        Predicts the output of a given model for a given set of inputs.

        Parameters:
        -----------
        timestamps : list
            A list of timestamps for prediction.
        inputs : dict, optional
            Inputs to be used for prediction. The keys must be the Codes of the
            input variables, and the values must be lists of the same length as the timestamps.
        timestamps_unit : str, optional
            Unit of the timestamps, by default "s".
            Needs to be one of the following: "s", "m", "h", "d".
        config: PredictionConfig, optional
            Configuration for the prediction method. Refer to the PredictionConfig class for
            details on the configuration parameters.
            See also: `PredictionConfig`


        Returns:
        --------
        Dictionary with predictions where:
            key: variable code
            value: list with predictions for each spectrum

        Example:
        --------
        >>> timestamps = [1,2,3,4,5,6,7]
        >>> inputs = {"var1": [42], "var2": [0.3], "var3": [0.5], "var4": [0,2,3,3,3,3,3]}
        >>> result = model.predict(timestamps, inputs, timestamps_unit="d")
        >>> print(result)
        {'var2': [pred1, pred2, pred3, pred4, pred5, pred6, pred7],
        'var3': [pred1, pred2, pre3, pred4, pred5, pred6, pred7]}

        """

        if not self.success:
            raise ModelPredictionException(
                f"{self.name} is not ready for prediction. The current status is {self.status}"
            )

        prediction_config = PredictionRequestConfig.new(
            model_confidence=config.model_confidence,
            starting_index=config.starting_index,
        )

        data_processing_strategy = CultivationPropagationPreprocessor(
            timestamps=timestamps,
            timestamps_unit=timestamps_unit,
            inputs=inputs,
            prediction_config=prediction_config,
            model=self,
        )

        return super().get_predictions(data_processing_strategy)

    @staticmethod
    def requests(client: Client) -> CRUDClient["CultivationPropagationModel"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["CultivationPropagationModel"](
            client, MODELS_URL, CultivationPropagationModel
        )


class CultivationHistoricalModel(CultivationModel):
    """Pydantic Model for Historical Model for Cultivation from the API"""

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    # pylint: disable=arguments-renamed
    def predict(
        self,
        timestamps: list[Union[int, float]],
        steps: list[Optional[int]],
        inputs: dict[str, list],
        timestamps_unit: str = "s",
        config: PredictionConfig = PredictionConfig(),
    ) -> dict:
        """
        Predicts the output of a given model for a given set of inputs.

        Parameters:
        -----------
        timestamps : list
            A list of timestamps for prediction.
        steps : list
            A list of steps for prediction. This steps should match the length of the timestamps
            and start as 0, representing the steps from the start of the process.
        inputs : dict, optional
            Inputs to be used for prediction. The keys must be the Codes of the
            input variables, and the values must be lists of the same length as the timestamps.
        timestamps_unit : str, optional
            Unit of the timestamps, by default "s".
            Needs to be one of the following: "s", "m", "h", "d".
        config: PredictionConfig, optional
            Configuration for the prediction method. Refer to the PredictionConfig class for
            details on the configuration parameters.
            See also: `PredictionConfig`

        Returns:
        --------
        Dictionary with predictions where:
            key: variable code
            value: list with predictions for each spectrum

        Example:
        --------
        >>> timestamps = [86400, 172800, 259200, 345600, 432000, 518400, 604800]
        >>> steps = [0,1,2,3,4,5,6]
        >>> inputs = {"var1": [42], "var2": [0.3], "var3": [0.5], "var4": [0,2,3,3,3,3,3]}
        >>> result = model.predict(timestamps, steps, inputs, timestamps_unit="s")
        >>> print(result)
        {'output1': [pred1], 'output2': [pred2]}

        """

        if not self.success:
            raise ModelPredictionException(
                f"{self.name} is not ready for prediction. The current status is {self.status}"
            )

        prediction_config = PredictionRequestConfig.new(
            model_confidence=config.model_confidence
        )

        data_processing_strategy = CultivationHistoricalPreprocessor(
            timestamps=timestamps,
            timestamps_unit=timestamps_unit,
            steps=steps,
            inputs=inputs,
            prediction_config=prediction_config,
            model=self,
        )

        return super().get_predictions(data_processing_strategy)

    @staticmethod
    def requests(client: Client) -> CRUDClient["CultivationHistoricalModel"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["CultivationHistoricalModel"](
            client, MODELS_URL, CultivationHistoricalModel
        )


# pylint: disable=too-few-public-methods
class ModelFactory:
    """Factory for Model, given the process unit id and model type"""

    MODEL_MAP = {
        "373c173a-1f23-4e56-874e-90ca4702ec0d": SpectraModel,
        "04a324da-13a5-470b-94a1-bda6ac87bb86": CultivationModel,
    }

    def __init__(self, process_unit_id):
        self._process_unit_id = process_unit_id

    def get_model(self, **kwargs) -> Type[Model]:
        """Get the model type from the process unit id"""

        if self._process_unit_id not in self.MODEL_MAP:
            raise NotImplementedError(
                f"Process unit id {self._process_unit_id} is not currently supported"
            )

        model = self.MODEL_MAP[self._process_unit_id]

        if "model_type" in kwargs and model == CultivationModel:
            if kwargs["model_type"] == "propagation":
                model = CultivationPropagationModel
            elif kwargs["model_type"] == "historical":
                model = CultivationHistoricalModel

        return model


class Project(BaseModel, ABC):
    """Abstract class for a DHL Project"""

    id: str = Field(alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")
    process_unit_id: str = Field(alias="processUnitId")
    _client: Client = PrivateAttr()

    def get_datasets(self, name: Optional[str] = None) -> Result[Dataset]:
        """Get the datasets of the project from the API

        Parameters:
        -----------
        name: str, optional
            Name of the dataset to be retrieved, by default None

        Returns:
        --------

        Result[Dataset]: Iteratable object containing the that correspond to the search

        """

        datasets = Dataset.requests(self._client)
        query_params = {
            "filterBy[projectId]": self.id,
        }

        if name is not None:
            query_params.update({"filterBy[name]": name})

        results = Result[Dataset](
            limit=5,
            query_params=query_params,
            requests=datasets,
        )

        return results

    @abstractmethod
    def get_models(self, name: Optional[str] = None, **kwargs) -> Result[Model]:
        """Get the models of the project from the API

        Parameters:
        -----------
        name: str, optional
            Name of the model to be retrieved, by default None
        model_type: str, optional
            For Cultivation Models, the type of the model to be retrieved,
            it can be `propagation` or `historical`, by default `propagation`
        """

    @abstractmethod
    def _get_model_query_params(self, name: Optional[str] = None) -> dict[str, str]:
        """Get the query params for the models"""

    @staticmethod
    @abstractmethod
    def requests(client: Client) -> CRUDClient["Project"]:
        """Resquests abstract method for Project Types"""


class SpectraProject(Project):
    """Pydantic Model for a DHL Project from the API"""

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    def get_models(self, name: Optional[str] = None) -> Result[Model]:
        """Get the models of the project from the API"""

        model = ModelFactory(self.process_unit_id).get_model()

        models = model.requests(self._client)
        query_params = self._get_model_query_params(name=name)

        results = Result[model](
            limit=5,
            query_params=query_params,
            requests=models,
        )

        return results

    def _get_model_query_params(self, name: Optional[str] = None) -> dict[str, str]:
        query_params = {"filterBy[projectId]": self.id}

        if name is not None:
            query_params.update({"filterBy[name]": name})

        return query_params

    @staticmethod
    def requests(client: Client) -> CRUDClient["SpectraProject"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["SpectraProject"](client, PROJECTS_URL, SpectraProject)


class CultivationProject(Project):
    """Pydantic Model for a DHL Project from the API"""

    id: str = Field(alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")
    process_unit_id: str = Field(alias="processUnitId")
    _client: Client = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._client = data["client"]

    def get_models(
        self,
        name: Optional[str] = None,
        model_type: Literal["propagation", "historical"] = "propagation",
    ) -> Result[Model]:
        """Get the models of the project from the API"""

        if model_type not in ["propagation", "historical"]:
            raise ValueError(
                f"model_type must be either propagation or historical, got {model_type}"
            )

        query_params = self._get_model_query_params(name=name, model_type=model_type)

        model = ModelFactory(self.process_unit_id).get_model(model_type=model_type)
        models = model.requests(self._client)

        results = Result[model](
            limit=5,
            query_params=query_params,
            requests=models,
        )
        return results

    def _get_model_query_params(
        self,
        name: Optional[str] = None,
        model_type: Literal["propagation", "historical"] = "propagation",
    ) -> dict[str, str]:
        query_params = {"filterBy[projectId]": self.id}

        if name is not None:
            query_params.update({"filterBy[name]": name})

        # get templateIds for propagation models
        template_query_params = {
            "filterByTag[type]": model_type,
            "archived": "any",
        }
        template_list = self._client.get(TEMPLATES_URL, template_query_params).json()
        template_ids = get_id_list(template_list)

        query_params.update({"filterBy[templateId]": "|".join(template_ids)})

        return query_params

    @staticmethod
    def requests(client: Client) -> CRUDClient["CultivationProject"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["CultivationProject"](
            client, PROJECTS_URL, CultivationProject
        )



"""This module contains generic utility functions used in the SDK
"""

import urllib.parse as urlparse
from datetime import datetime
from functools import reduce
from typing import Literal, Optional, Union

import numpy as np
from pydantic import BaseModel, Field, model_validator

from dhl_sdk._constants import GROUPS_URL
from dhl_sdk.crud import Client
from dhl_sdk.exceptions import InvalidConfidenceException

Predictions = dict[str, dict[str, list[float]]]


class VariableGroupCodes:
    """Singleton class to store the variable group codes"""

    _instance = None
    variable_group_codes = {}

    def __new__(cls, client):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialize(client)
        return cls._instance

    def _initialize(self, client: Client):
        group_codes = {}
        groups = client.get(GROUPS_URL).json()

        for group in groups:
            group_codes[group["name"]] = (group["id"], group["code"])

        self.variable_group_codes = group_codes

    def get_variable_group_codes(self) -> dict[str, tuple[str, str]]:
        """Returns the variable group codes saved in singleton"""
        return self.variable_group_codes


class Instance(BaseModel):
    """Pydantic class representing the Instance
    Used to type check the request"""

    timestamps: Optional[list[int]] = Field(default=None, alias="timestamps")
    sample_id: Optional[list[str]] = Field(default=None, alias="sampleId")
    steps: Optional[list[int]] = Field(default=None, alias="steps")
    values: Union[
        list[Optional[float]],
        list[Optional[bool]],
        list[Optional[str]],
        list[Optional[list[float]]],
    ]

    high_values: Optional[
        Union[list[Optional[float]], list[list[Optional[float]]]]
    ] = Field(default=None, alias="highValues")
    low_values: Optional[
        Union[list[Optional[float]], list[list[Optional[float]]]]
    ] = Field(default=None, alias="lowValues")

    @model_validator(mode="after")
    def generate_sample_ids(self):
        """Generates sample ids if not provided"""
        if self.sample_id is None:
            self.sample_id = [str(i) for i in range(len(self.values))]
        return self


class PredictionRequestConfig(BaseModel):
    """Prediction configurations"""

    starting_index: int = Field(alias="startingIndex", default=0)
    high_values_percentile: float = Field(alias="highValuesPercentile", default=90)
    low_values_percentile: float = Field(alias="lowValuesPercentile", default=10)

    @staticmethod
    def new(
        model_confidence: float, starting_index: int = 0
    ) -> "PredictionRequestConfig":
        """Create a new Prediction Configuration object"""
        if not 1.0 < model_confidence < 99.0:
            raise InvalidConfidenceException()

        high_value = 50.0 + model_confidence / 2
        low_value = 50.0 - model_confidence / 2

        return PredictionRequestConfig(
            startingIndex=starting_index,
            highValuesPercentile=high_value,
            lowValuesPercentile=low_value,
        )


class SpectraPredictionConfig(BaseModel):
    """Pydantic class representing Spectra Prediction Config"""

    prediction_mode: Literal["classic", "onlySpectra"] = Field(
        default="classic", alias="predictionMode"
    )


class OnlyId(BaseModel):
    """Pydantic class representing a sctuc with only the id"""

    id: str


class PredictionRequest(BaseModel):
    """Pydantic class representing the expected Predict Request"""

    instances: list[list[Optional[Instance]]]
    metadata: Optional[dict] = None
    config: Optional[Union[PredictionRequestConfig, SpectraPredictionConfig]] = None


class Metadata(BaseModel):
    """Pydantic class representing Metadata for Predict Request"""

    experiments: list[Optional[OnlyId]] = [None]
    variables: list[OnlyId]


class PipelineStage(BaseModel):
    """Pydantic class representing the Prediction Pipeline Stage"""

    config: Union[PredictionRequestConfig, SpectraPredictionConfig]
    id: str
    merge_strategy: str = Field(default="merge", alias="mergeStrategy")
    type: str = Field(default="predict")


class PredictionPipelineRequest(BaseModel):
    """Pydantic class representing the expected Predict Request"""

    instances: list[list[Optional[Instance]]]
    metadata: Metadata
    stages: list[PipelineStage]


class PredictionResponse(BaseModel):
    """Pydantic class representing the expected Predict Response"""

    instances: list[list[Optional[Instance]]]


def urljoin(*args) -> str:
    """join url elements together into one url"""
    elements = [
        f"{arg}/" if arg[-1] != "/" and i != len(args) - 1 else arg
        for i, arg in enumerate(args)
    ]
    return reduce(urlparse.urljoin, elements)


def validate_list_elements(arr: list) -> bool:
    """Validates if an array contains non float values"""
    return any(
        not isinstance(item, (float, int))
        or item is None
        or np.isnan(item)
        or np.isinf(item)
        for item in arr
    )


def get_id_list(json_list: list[dict]) -> list[str]:
    """Extracts the id from a list of dictionaries"""
    return [item["id"] for item in json_list]


def is_date_in_format(date_string, date_format):
    """Checks if a date string is in a specific format"""
    try:
        datetime.strptime(date_string, date_format)
        return True
    except ValueError:
        return False



# pylint: disable=too-few-public-methods

"""Client for the DHL SpectraHow API

This module defines the `Client` and `DataHowLabClient` classes, 
which are used to interact with the DataHowLab's API.

Classes:
    - Client: provides a base implementation for making HTTP requests to the API
    - DataHowLabClient: main client to interact with the DHL API
"""

from typing import Any, Dict, Literal, Optional, Type, TypeVar, Union
from urllib.parse import urlencode

import requests
from requests import Response
from requests.adapters import HTTPAdapter
import urllib3
from urllib3.util.retry import Retry

from dhl_sdk._constants import PROCESS_UNIT_MAP
from dhl_sdk._utils import VariableGroupCodes, urljoin
from dhl_sdk.authentication import APIKeyAuthentication
from dhl_sdk.crud import Result
from dhl_sdk.db_entities import DataBaseEntity, Experiment, Product, Recipe
from dhl_sdk.entities import CultivationProject, Project, SpectraProject, Variable

PROJECT_TYPE_MAP = {
    "04a324da-13a5-470b-94a1-bda6ac87bb86": CultivationProject,
    "373c173a-1f23-4e56-874e-90ca4702ec0d": SpectraProject,
}


T = TypeVar("T", bound=Project)


class Client:
    """
    A client for interacting with the DataHowLab API.
    """

    def __init__(
        self, auth_key: APIKeyAuthentication, base_url: str, verify: bool = True
    ) -> None:
        """
        Parameters
        ----------
        auth_key : APIKeyAuthentication
            An instance of the APIKeyAuthentication class containing the user's API key.
        base_url : str
            The URL address of the datahowlab application

        Returns
        -------
        NoneType
            None
        """
        self.auth_key = auth_key
        self.base_url = base_url
        self.session = Client._get_retry_requester(
            total_retries=5, backoff_factor=1, verify=verify
        )

    @staticmethod
    def _get_retry_requester(
        total_retries: int = 5, backoff_factor: int = 1, verify: int = True
    ):
        """Get the http session with retry strategy"""
        status_forcelist = [429, 502, 503, 504]
        allowed_methods = ["DELETE", "GET", "HEAD", "OPTIONS", "PUT", "TRACE"]

        retry_strategy = Retry(
            total=total_retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            allowed_methods=allowed_methods,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.verify = verify
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        return http

    def post(self, path: str, json_data: Any) -> Response:
        """
        Sends a POST request to the specified
        URL with the provided JSON data as a dict.

        Parameters
        ----------
        path : str
            The extension to send the POST request to.
        json_data : any
            The JSON data to include in the POST request.

        Returns
        -------
        requests.Response
            The response object returned by the server.

        Raises
        ------
        requests.exceptions.HTTPError
            If the server returns a non-2xx status code.
        """
        path = urljoin(self.base_url, path)
        auth_headers = self.auth_key.get_headers()

        response = self.session.post(path, headers=auth_headers, json=json_data)
        response.raise_for_status()

        return response

    def get(self, path: str, query_params: Optional[Dict[str, str]] = None) -> Response:
        """
        Sends a GET request to the specified URL
        with the given query parameters.

        Parameters:
        -----------
        path : str
            The URL to send the GET request to.
        query_params : dict, optional
            A dictionary of query parameters to include in the request.

        Returns:
        --------
        response : requests.Response
            The response object returned by the GET request.

        Raises:
        -------
        requests.exceptions.HTTPError
            If the server returns a non-2xx status code.
        """
        if query_params:
            query_string = urlencode(query_params, doseq=True, safe="[]")
            path = f"{path}?{query_string}"

        path = urljoin(self.base_url, path)
        auth_headers = self.auth_key.get_headers()

        response = self.session.get(path, headers=auth_headers)
        response.raise_for_status()

        return response

    def put(
        self,
        path: str,
        data: Any,
        content_type: Literal["application/json", "text/csv"] = "application/json",
    ) -> Response:
        """
        Sends a PUT request to the specified
        URL with the provided JSON data as a dict.

        Parameters
        ----------
        path : str
            The extension to send the PUT request to.
        data : any
            The data to include in the PUT request.
        content_type : Literal["application/json", "text/csv"], optional
            The content type of the data to be sent, by default "application/json"

        Returns
        -------
        requests.Response
            The response object returned by the server.

        Raises
        ------
        requests.exceptions.HTTPError
            If the server returns a non-2xx status code.
        """
        path = urljoin(self.base_url, path)
        req_headers = self.auth_key.get_headers()

        if content_type == "application/json":
            response = self.session.put(path, headers=req_headers, json=data)
            response.raise_for_status()

        elif content_type == "text/csv":
            req_headers["Content-type"] = content_type
            response = self.session.put(path, headers=req_headers, data=data)
            response.raise_for_status()
        else:
            raise NotImplementedError(
                f"Put request with given content type '{content_type}' not implemented yet."
            )

        return response

    def get_projects(
        self,
        project_type: Type[T],
        name: Optional[str] = None,
        unit_id: Optional[str] = None,
        offset: int = 0,
    ) -> Result[T]:
        """Retrieve the available projects for the user

        Parameters
        ----------
        project_type : T, optional
            The type of project to retrieve, by default Project
        name : str, optional
            Filter projects by name, by default None
        unit_id : str, optional
            Filter projects by process unit ID, by default None
        offset : int, optional
            The offset for pagination, must be a non-negative integer, by default 0

        Returns
        -------
        Result
            An Iterable object containing the retrieved project data
        """

        if not (isinstance(offset, int) and offset >= 0):
            raise ValueError("offset must be a non-negative integer")

        filter_params = {
            key: value
            for key, value in {
                "filterBy[processUnitId]": unit_id,
                "filterBy[name]": name,
            }.items()
            if value is not None
        }

        projects = project_type.requests(self)

        result = Result[project_type](
            offset=offset,
            limit=10,
            query_params=filter_params,
            requests=projects,
        )

        return result


class DataHowLabClient:
    """

    Client for the DHL API

    """

    def __init__(
        self,
        auth_key: APIKeyAuthentication,
        base_url: str,
        verify_ssl: Union[bool, str] = True,
    ):
        """
        Parameters
        ----------
        auth_key : APIKeyAuthentication
            An instance of the APIKeyAuthentication class containing the user's API key.
        base_url : str
            The URL address of the datahowlab application
        verify_ssl : Union[bool, str], optional
            Either a boolean, in which case it controls whether we verify the server's
            TLS certificate, or a string, in which case it must be a path to a CA bundle
            to use. For more info check the documentation for python's requests.request.
            By default True.

        Returns
        -------
        NoneType
            None
        """

        if isinstance(verify_ssl, bool) and not verify_ssl:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self._client = Client(auth_key, base_url, verify=verify_ssl)

    def get_projects(
        self,
        name: Optional[str] = None,
        project_type: Literal["cultivation", "spectroscopy"] = "cultivation",
    ) -> Result[Project]:
        """
        Retrieves an iterable of projects from the DHL API.

        Parameters
        ----------
        name : str, optional
            A string to filter projects by name.
        offset : int, optional
            An integer representing the number of projects to skip before returning results.
        project_type : Literal["cultivation", "spectroscopy"], optional
            The type of project to retrieve, by default 'cultivation'

        Returns
        -------
        Result
            An Iterable object containing the retrieved projects
        """

        if project_type not in PROCESS_UNIT_MAP:
            raise ValueError(
                f"Type must be one of {list(PROCESS_UNIT_MAP.keys())}, "
                "but got '{project_type}'"
            )

        unit_id = PROCESS_UNIT_MAP[project_type]

        project_class = PROJECT_TYPE_MAP[unit_id]

        return self._client.get_projects(
            name=name,
            unit_id=unit_id,
            project_type=project_class,
        )

    def get_experiments(
        self,
        name: Optional[str] = None,
        product: Optional[Product] = None,
    ) -> Result[Experiment]:
        """Retrieve the available experiments for the user

        Parameters
        ----------
        name : str, optional
            Search in DB by name, by default None
        product : Product, optional
            Filter experiments by product, by default None

        Returns
        -------
        Result
            An Iterable object containing the retrieved experiment data
        """

        product_id = product.id if product else None

        filter_params = {
            key: value
            for key, value in {
                "search": name,
                "filterBy[product._id]": product_id,
            }.items()
            if value is not None
        }

        experiments = Experiment.requests(self._client)
        result = Result[Experiment](
            offset=0,
            limit=10,
            query_params=filter_params,
            requests=experiments,
        )

        return result

    def get_products(self, code: Optional[str] = None) -> Result[Product]:
        """Retrieve the available products for the user

        Parameters
        ----------
        code : str, optional
            Filter products by code, by default None

        Returns
        -------
        Result
            An Iterable object containing the retrieved product data
        """

        filter_params = {"filterBy[code]": code} if code else None

        projects = Product.requests(self._client)
        result = Result[Product](
            offset=0,
            limit=10,
            query_params=filter_params,
            requests=projects,
        )

        return result

    def get_variables(
        self,
        code: Optional[str] = None,
        group: Optional[str] = None,
        variable_type: Optional[
            Literal["categorical", "flow", "logical", "numeric"]
        ] = None,
    ) -> Result[Variable]:
        """Retrieve the available variables for the user

        Parameters
        ----------
        code : str, optional
            Filter variables by code, by default None

        Returns
        -------
        Result
            An Iterable object containing the retrieved variable data
        """

        if variable_type and variable_type not in [
            "categorical",
            "flow",
            "logical",
            "numeric",
        ]:
            raise ValueError(
                (
                    f"Variable Type must be one of: 'categorical', 'flow',"
                    f" 'logical', 'numeric'. instead, it got '{variable_type}'"
                )
            )

        if group:
            variable_group_codes = VariableGroupCodes(
                self._client
            ).get_variable_group_codes()

            if group not in variable_group_codes:
                raise ValueError(
                    f"Variable Group must be one of: {list(variable_group_codes.keys())}."
                    f" instead, it got '{group}'"
                )

            group_id = variable_group_codes[group][0]
        else:
            group_id = None

        filter_params = {
            key: value
            for key, value in {
                "filterBy[code]": code,
                "filterBy[variant]": variable_type,
                "filterBy[group._id]": group_id,
            }.items()
            if value is not None
        }

        projects = Variable.requests(self._client)
        result = Result[Variable](
            offset=0,
            limit=10,
            query_params=filter_params,
            requests=projects,
        )

        return result

    def get_recipes(
        self, name: Optional[str] = None, product: Optional[Product] = None
    ) -> Result[Recipe]:
        """Retrieve the available recipes for the user

        Parameters
        ----------
        name : str, optional
            Filter recipes by name, by default None
        product : Product, optional
            Filter recipes by product, by default None

        Returns
        -------
        Result
            An Iterable object containing the retrieved recipe data
        """

        product_id = product.id if product else None

        filter_params = {
            key: value
            for key, value in {
                "filterBy[name]": name,
                "filterBy[product._id]": product_id,
            }.items()
            if value is not None
        }

        recipes = Recipe.requests(self._client)
        result = Result[Recipe](
            offset=0,
            limit=10,
            query_params=filter_params,
            requests=recipes,
        )

        return result

    def create(self, entity: DataBaseEntity) -> Optional[DataBaseEntity]:
        """Create a new entity in the database

        Parameters
        ----------
        entity : DataBaseEntity
            The entity to create in the database.
            This can be a Product, Variable, Recipe or Experiment.

        Returns
        -------
        DataBaseEntity
            The created entity
        """

        if entity.validate_import(self._client):
            return entity.requests(self._client).create(entity.create_request_body())
        else:
            return entity



""" New Entity Validators"""

# pylint: disable=missing-function-docstring, arguments-differ
# pylint: disable=missing-class-docstring, protected-access

from datetime import datetime
import math
import warnings
from abc import ABC, abstractmethod
from typing import Any, Optional, Protocol, Union

from dhl_sdk._input_processing import (
    groupcode_is_output,
    groupcode_is_timedependent,
    variant_is_numeric,
)
from dhl_sdk._utils import is_date_in_format
from dhl_sdk._constants import (
    EXPERIMENTS_URL,
    FILES_URL,
    PRODUCTS_URL,
    VARIABLES_URL,
)
from dhl_sdk.crud import Client, CRUDClient
from dhl_sdk.exceptions import ImportValidationException


class AbstractValidator(ABC):
    """Abstract class for Validators"""

    @abstractmethod
    def validate(self, **kwargs) -> bool:
        """Validate the entity for importing"""

    @abstractmethod
    def is_imported(self, **kwargs) -> bool:
        """Check if the entity is already imported"""


class Group(Protocol):
    id: str
    name: str
    code: str


class FlowReferences(Protocol):
    measurement_id: str
    concentration_id: Optional[str]
    fraction_id: Optional[str]


class FlowVariableDetails(Protocol):
    references: list[FlowReferences]


class Variable(Protocol):
    id: str
    code: str
    name: str
    group: Group
    variant: str
    variant_details: Optional[Union[Any, FlowVariableDetails]]
    measurement_unit: Optional[str]
    _validator: AbstractValidator


class Product(Protocol):
    id: str
    code: str
    name: str
    _validator: AbstractValidator


class File(Protocol):
    id: str
    name: str
    type: str


class Experiment(Protocol):
    id: str
    name: str
    product: Product
    variables: list[Variable]
    file_data: Any
    _validator: AbstractValidator

    @staticmethod
    def requests(client: Client) -> CRUDClient["Experiment"]:
        # pylint: disable=missing-function-docstring
        ...


class Recipe(Protocol):
    id: str
    name: str
    product: Product
    variables: list[Variable]
    file_data: Any
    _validator: AbstractValidator

    @staticmethod
    def requests(client: Client) -> CRUDClient["Recipe"]:
        # pylint: disable=missing-function-docstring
        ...


class VariableValidator(AbstractValidator):
    """Validator for Variables"""

    def validate(self, entity: Variable, client: Client) -> bool:
        """Validate the variable for importing"""

        # remove whitespaces from code
        entity.code = entity.code.strip()

        # first validate if variable is alredy in the DB
        if self.is_imported(entity, client):
            warnings.simplefilter("always")
            warnings.warn(
                f"Variable {entity.code} is already present in the DB. Skipping import."
            )
            return False

        validation_errors = []

        # validate if variable code already exists
        query_params = {
            "filterBy[code]": entity.code,
            "filterBy[group._id]": entity.group.id,
            "archived": "any",
        }

        response = client.get(VARIABLES_URL, query_params=query_params)
        if int(response.headers.get("x-total-count")) > 0:  # type: ignore
            validation_errors.append(f"The variable code {entity.code} already exists")

        # Validate if variable name already exists
        query_params = {
            "filterBy[name]": entity.name,
            "filterBy[group._id]": entity.group.id,
            "archived": "any",
        }

        response = client.get(VARIABLES_URL, query_params=query_params)
        if int(response.headers.get("x-total-count")) > 0:  # type: ignore
            validation_errors.append(f"The variable name {entity.name} already exists")

        # Validate if measurement unit is present
        if entity.measurement_unit is None:
            validation_errors.append("The measurement unit must be present")

        # missing variant validation
        if entity.variant == "flow":
            references = entity.variant_details.references
            if references is None:
                validation_errors.append(
                    "Flow variables must have a list of references"
                )

            for reference in references:
                if not reference.measurement_id:
                    validation_errors.append(
                        "Variable Flow references must have"
                        " a reference to a X variable (measurement_id)"
                    )
                else:
                    measurement_id = reference.measurement_id
                    response = client.get(f"{VARIABLES_URL}/{measurement_id}")
                    if response.status_code != 200:
                        validation_errors.append(
                            f"Variable with id {measurement_id} does not exist."
                            " Please make sure you are using a variable that is"
                            " already in the database"
                        )
                    else:
                        measurement = response.json()
                        if measurement["group"]["code"] != "X":
                            validation_errors.append(
                                f"Variable with id {measurement_id} must be a "
                                "X variable to be valid as a reference"
                            )

                if reference.concentration_id:
                    concentration_id = reference.concentration_id
                    response = client.get(f"{VARIABLES_URL}/{concentration_id}")
                    if response.status_code != 200:
                        validation_errors.append(
                            f"Variable with id {concentration_id} does not exist."
                            " Please make sure you are using a variable that is "
                            "already in the database"
                        )
                    else:
                        concentration = response.json()
                        if concentration["group"]["code"] != "FeedConc":
                            validation_errors.append(
                                f"Variable with id {concentration_id} must be a "
                                "Feed Concentration variable to be valid as a reference"
                            )

                if reference.fraction_id:
                    fraction_id = reference.fraction_id
                    response = client.get(f"{VARIABLES_URL}/{fraction_id}")
                    if response.status_code != 200:
                        validation_errors.append(
                            f"Variable with id {fraction_id} does not exist."
                            " Please make sure you are using a variable that is "
                            "already in the database"
                        )

        if validation_errors:
            raise ImportValidationException("\n".join(validation_errors))

        return True

    def is_imported(self, entity: Variable, client: Client) -> bool:
        """Check if the variable is already imported"""

        # check if variable has no id. If it has no id, check if there is already a variable
        # with the same code, name and group in the DB. If so, set the id of the variable
        # and return True

        if not entity.id:
            query_params = {
                "filterBy[code]": entity.code,
                "filterBy[name]": entity.name,
                "filterBy[measurementUnit]": entity.measurement_unit,
                "filterBy[group._id]": entity.group.id,
                "archived": "any",
            }

            response = client.get(VARIABLES_URL, query_params=query_params)
            if int(response.headers.get("x-total-count")) > 0:
                entity.id = response.json()[0]["id"]
                return True
            else:
                return False

        response = client.get(f"{VARIABLES_URL}/{entity.id}")
        return response.status_code == 200


class ProductValidator(AbstractValidator):
    """Validator for Products"""

    def validate(self, entity: Product, client: Client) -> bool:
        validation_errors = []

        # remove whitespaces from code
        entity.code = entity.code.strip()

        if self.is_imported(entity, client):
            warnings.simplefilter("always")
            warnings.warn(
                f"Product {entity.code} is already present in the DB. Skipping import."
            )
            return False

        # check if length of code is less than 5
        if len(entity.code) > 6:
            raise ImportValidationException(
                "Product code must be from 1 to 6 characters long"
            )

        # validate if variable code already exists
        query_params = {
            "filterBy[code]": entity.code,
            "archived": "any",
        }

        response = client.get(PRODUCTS_URL, query_params=query_params)
        if int(response.headers.get("x-total-count")) > 0:
            validation_errors.append(
                f"This product code {entity.code} is already taken"
            )

        # Validate if variable name already exists
        query_params = {
            "filterBy[name]": entity.name,
            "archived": "any",
        }

        response = client.get(PRODUCTS_URL, query_params=query_params)
        if int(response.headers.get("x-total-count")) > 0:
            validation_errors.append(
                f"This product name {entity.name} is already taken"
            )

        if validation_errors:
            raise ImportValidationException("\n".join(validation_errors))

        return True

    def is_imported(self, entity: Product, client: Client) -> bool:
        """Check if the variable is already imported"""

        if not entity.id:
            query_params = {
                "filterBy[code]": entity.code,
                "filterBy[name]": entity.name,
                "archived": "any",
            }

            response = client.get(PRODUCTS_URL, query_params=query_params)
            if int(response.headers.get("x-total-count")) > 0:
                entity.id = response.json()[0]["id"]
                return True
            else:
                return False

        response = client.get(f"{PRODUCTS_URL}/{entity.id}")
        return response.status_code == 200


class AbstractFileValidator(ABC):
    """Abstract class for File Validators"""

    @classmethod
    @abstractmethod
    def format_data(
        cls,
        variables: list[Variable],
        data: Any,
    ) -> Any:
        """Format the file data"""

    @abstractmethod
    def validate(self, variables: list[Variable], data: Any, **kwargs) -> bool:
        """Validate the file for importing"""

    @abstractmethod
    def is_imported(self, entity: File, client: Client) -> bool:
        """Check if the file is already imported"""


class RecipeFileValidator(AbstractFileValidator):
    """Validator for Files"""

    def __init__(self, duration: Optional[int] = None) -> None:
        super().__init__()
        self.duration = duration

    @classmethod
    def format_data(cls, variables: list[Variable], data: Any) -> Any:
        """Format the file data"""

        # only store the first value of X Variables
        for variable in variables:
            if variable.group.code == "X" and variable.code in data:
                if len(data[variable.code]["timestamps"]) > 1:
                    warnings.warn(
                        f"Variable {variable.code} has more than 1 timestamp."
                        " Only the first value will be stored"
                    )

                    data[variable.code]["timestamps"] = [
                        data[variable.code]["timestamps"][0]
                    ]
                    data[variable.code]["values"] = [data[variable.code]["values"][0]]

            if variable.variant == "categorical":
                data[variable.code]["values"] = [
                    str(value) for value in data[variable.code]["values"]
                ]

        return data

    def validate(self, variables: list[Variable], data: Any) -> bool:
        """Validate the file for importing"""
        validation_errors = []

        # check if there are X variables
        if not any(var.group.code == "X" for var in variables):
            raise ImportValidationException(
                (
                    "No variables 'X' found. 'X' variables"
                    " are mandatory for the recipe."
                )
            )

        timedependent_variable_size = None

        for variable in variables:
            if variable.code in data:
                # check if there are output varlues
                if groupcode_is_output(variable.group.code):
                    validation_errors.append(
                        f"Variable {variable.code} is an Output,"
                        " so it's not allowed in the recipe"
                    )
                    continue

                # check if , for feedConc variables, the corresponding feed
                # and X variables are present
                if variable.group.code == "FeedConc":
                    var_code = variable.code
                    feedconc_codes = var_code.split("_")
                    if len(feedconc_codes) != 2:
                        validation_errors.append(
                            f"Variable {variable.code} must have the format 'Feed_XVar'"
                        )
                        continue

                    if feedconc_codes[0] not in data.keys():
                        validation_errors.append(
                            f"Variable {variable.code} does not have a corresponding Feed Variable"
                        )
                        continue

                    if feedconc_codes[1] not in data.keys():
                        validation_errors.append(
                            f"Variable {variable.code} does not have a corresponding X Variable"
                        )
                        continue

                variable_data = data[variable.code]
                if not isinstance(variable_data, dict):
                    validation_errors.append(
                        (
                            f"Variable {variable.code} data must be a dictionary with"
                            "the mandatory fields 'timestamps' and 'values'"
                        )
                    )
                    continue

                if "timestamps" not in variable_data or "values" not in variable_data:
                    validation_errors.append(
                        f"Variable {variable.code} data must contain 'timestamps' and 'values' key"
                    )
                    continue

                timestamps = variable_data["timestamps"]
                values = variable_data["values"]

                if not isinstance(timestamps, list) or not isinstance(values, list):
                    validation_errors.append(
                        (
                            f"Variable {variable.code} data must "
                            " contain 'timestamps' and 'values' as lists"
                        )
                    )
                    continue

                if len(values) != len(timestamps) or len(timestamps) == 0:
                    validation_errors.append(
                        (
                            f"Variable {variable.code} data can't "
                            " be empty and must have equal length "
                            "of timestamps and values"
                        )
                    )
                    continue

                if None in values:
                    validation_errors.append(
                        (
                            f"Variable {variable.code} contains None values, that are not allowed"
                        )
                    )
                    continue

                if groupcode_is_timedependent(variable.group.code):
                    if len(timestamps) < 2:
                        validation_errors.append(
                            (
                                f"Variable {variable.code} must have the"
                                " more than 1 timestamp"
                            )
                        )
                        continue

                    if timedependent_variable_size is None:
                        timedependent_variable_size = len(timestamps)
                    else:
                        if len(timestamps) != timedependent_variable_size:
                            validation_errors.append(
                                (
                                    f"Variable {variable.code} must have the"
                                    " same length as the other time "
                                    " dependent variables {timedependent_variable_size}"
                                )
                            )
                            continue

                    if self.duration:
                        if (timestamps[-1] - timestamps[0]) != self.duration:
                            validation_errors.append(
                                (
                                    f"Variable {variable.code} timestamps must"
                                    " have the duration of the recipe"
                                )
                            )
                            continue

                else:
                    # validate if non time dependent inputs have length of 1 (only initial values)
                    if len(values) != 1 and variable.group.code != "X":
                        validation_errors.append(
                            (
                                f"Input {variable.code} only requires initial "
                                f"values, so it must have a length of 1"
                            )
                        )
                        continue

                if variant_is_numeric(variable.variant):
                    if not all(
                        isinstance(value, (int, float)) and math.isfinite(value)
                        for value in values
                    ):
                        validation_errors.append(
                            (f"Variable {variable.code} values must be numeric")
                        )
                        continue

            else:
                if groupcode_is_output(variable.group.code):
                    continue

                validation_errors.append(
                    f"Variable {variable.code} is missing from the data dictionary"
                )

        if validation_errors:
            raise ImportValidationException("\n".join(validation_errors))

        return True

    def is_imported(self, entity: File, client: Client) -> bool:
        """Check if the file is already imported"""

        if not entity.id:
            return False

        response = client.get(f"{FILES_URL}/{entity.id}")
        return response.status_code == 200


class ExperimentFileValidator(AbstractFileValidator):
    """Validator for Files"""

    @classmethod
    def format_data(cls, variables: list[Variable], data: Any) -> Any:
        """Format the file data"""

        for variable in variables:
            if variable.variant == "categorical":
                data[variable.code]["values"] = [
                    str(value) for value in data[variable.code]["values"]
                ]

        return data

    def validate(
        self,
        variables: list[Variable],
        data: Any,
        variant: str = "run",
        variant_details: Optional[dict] = None,
    ) -> bool:
        """Validate the file for importing"""
        validation_errors = []

        if variant == "run":
            sample_id = "timestamps"
        elif variant == "samples":
            sample_id = "sampleId"

        for variable in variables:
            if variable.code.lower() not in (key.lower() for key in data):
                validation_errors.append(
                    f"Variable {variable.code} is missing from the data dictionary"
                )
                continue

            variable_data = data[variable.code]
            if not isinstance(variable_data, dict):
                validation_errors.append(
                    (
                        f"Variable {variable.code} data must be a dictionary with"
                        f"the mandatory fields {sample_id} and 'values'"
                    )
                )
                continue

            if sample_id not in variable_data or "values" not in variable_data:
                validation_errors.append(
                    f"Variable {variable.code} data must contain {sample_id} and 'values' key"
                )
                continue

            ids = variable_data[sample_id]
            values = variable_data["values"]

            if not isinstance(ids, list) or not isinstance(values, list):
                validation_errors.append(
                    f"Variable {variable.code} data must contain {sample_id} and 'values' as lists"
                )
                continue

            if len(ids) != len(values) or len(ids) == 0:
                validation_errors.append(
                    (
                        f"Variable {variable.code} data can't be empty and must have equal length"
                        f"of {sample_id} and values"
                    )
                )
                continue

            if variant_is_numeric(variable.variant):
                if not all(isinstance(value, (int, float)) for value in values):
                    validation_errors.append(
                        f"Values must be numeric for variable {variable.code}"
                        f" of variant {variable.variant}"
                    )
                    continue

            elif variable.variant == "logical":
                if not all(isinstance(value, bool) for value in values):
                    validation_errors.append(
                        f"Values must be boolean for variable {variable.code}"
                        f" of variant {variable.variant}"
                    )
                    continue

            # validations only for run variant
            if sample_id == "timestamps":
                # check if timestamps are sorted
                if not all(ids[i] < ids[i + 1] for i in range(len(ids) - 1)):
                    validation_errors.append(
                        f"Variable {variable.code} timestamps must be sorted"
                    )
                    continue

                if not variant_details:
                    validation_errors.append(
                        "Start and end time are required for the experiment import."
                    )
                    continue

                start_time = variant_details.get("startTime", None)
                end_time = variant_details.get("endTime", None)

                if not start_time or not end_time:
                    validation_errors.append(
                        "Start and end time are required for the experiment import"
                    )
                    continue

                start_timestamp = int(
                    datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%SZ").timestamp()
                )
                end_timestamp = int(
                    datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%SZ").timestamp()
                )

                # check if all timestamps are within the start and end time
                if ids[0] < start_timestamp or ids[-1] > end_timestamp:
                    validation_errors.append(
                        f"Variable {variable.code} timestamps must be within the start and "
                        "end time of the experiment to be correctly tabularized. "
                        " Please modify the timestamps or the start and end time of the experiment"
                    )
                    continue

        if validation_errors:
            raise ImportValidationException("\n".join(validation_errors))

        return True

    def is_imported(self, entity: File, client: Client) -> bool:
        """Check if the file is already imported"""

        if not entity.id:
            return False

        response = client.get(f"{FILES_URL}/{entity.id}")
        return response.status_code == 200


class SpectraFileValidator(AbstractFileValidator):
    """Validator for Files"""

    @classmethod
    def format_data(cls, variables: list, data: Any) -> Any:
        """Format the file data"""
        return data

    def validate(
        self, variables: list[Variable], data: Any, variant: str = "run"
    ) -> bool:
        """Validate the file for importing"""
        validation_errors = []

        if variant == "run":
            sample_id = "timestamps"
        elif variant == "samples":
            sample_id = "sampleId"

        if "spectra" not in data:
            raise ImportValidationException(
                ("Spectra data is missing. Please add the 'spectra' key to the data")
            )

        # save spectra variable in variable list
        spectra_variable = None
        for variable in variables:
            if variable.group.code == "SPC":
                if spectra_variable is not None:
                    raise ImportValidationException(
                        "Only one variable of the group Spectra is allowed"
                    )
                spectra_variable = variable
                variable_code = "spectra"

            else:
                variable_code = variable.code
                if variable.code.lower() not in (key.lower() for key in data):
                    validation_errors.append(
                        f"Variable {variable.code} is missing from the data dictionary"
                    )
                    continue

            variable_data = data[variable_code]
            if not isinstance(variable_data, dict):
                validation_errors.append(
                    (
                        f"Variable {variable_code} data must be a dictionary with"
                        f"the mandatory fields {sample_id} and 'values'"
                    )
                )
                continue

            if sample_id not in variable_data or "values" not in variable_data:
                validation_errors.append(
                    f"Variable {variable_code} data must contain {sample_id} and 'values' key"
                )
                continue

            ids = variable_data[sample_id]
            values = variable_data["values"]

            if not isinstance(ids, list) or not isinstance(values, list):
                validation_errors.append(
                    f"Variable {variable_code} data must contain {sample_id} and 'values' as lists"
                )
                continue

            if len(ids) != len(values) or len(ids) == 0:
                validation_errors.append(
                    (
                        f"Variable {variable_code} data can't be empty and must have equal length"
                        f"of {sample_id} and values"
                    )
                )
                continue

            if variable_code == "spectra":
                if sample_id == "timestamps":
                    for date in ids:
                        if not is_date_in_format(date, "%Y-%m-%dT%H:%M:%S.%fZ"):
                            validation_errors.append(
                                f"Timestamp {date} is not in the correct format."
                                "The timestamp must be in the format %Y-%m-%dT%H:%M:%S.%fZ"
                            )
                            continue

                if not all(isinstance(value, list) for value in values):
                    validation_errors.append(
                        f"Variable {variable_code} values must be a list of lists"
                    )
                    continue

                spectra_size = variable.size
                for i, value in enumerate(values):
                    if len(value) != spectra_size:
                        validation_errors.append(
                            (
                                f"Spectrum of index {i} has the wrong dimensions."
                                f"It should be of length {spectra_size}"
                            )
                        )
                        continue

                    if not all(isinstance(val, (int, float)) for val in value):
                        validation_errors.append(
                            f"Variable {variable_code} values must all be numeric"
                        )
                        continue

        if spectra_variable is None:
            raise ImportValidationException(
                "A variable of the group Spectra is missing from the variable list"
            )

        if validation_errors:
            raise ImportValidationException("\n".join(validation_errors))

        return True

    def is_imported(self, entity: File, client: Client) -> bool:
        """Check if the file is already imported"""

        if not entity.id:
            return False

        response = client.get(f"{FILES_URL}/{entity.id}")
        return response.status_code == 200


class ExperimentValidator(AbstractValidator):
    """Validator for Experiments"""

    def validate(self, entity: Union[Experiment, Recipe], client: Client) -> bool:
        if self.is_imported(entity, client):
            warnings.simplefilter("always")
            warnings.warn(
                f"Experiment {entity.name} is already present in the DB. Skipping import."
            )
            return False

        # Validate if experiment name already exists
        query_params = {
            "filterBy[name]": entity.name,
            "archived": "false",
        }

        response = client.get(EXPERIMENTS_URL, query_params=query_params)
        if int(response.headers.get("x-total-count")) > 0:  # type: ignore
            raise ImportValidationException(
                f"The experiment name {entity.name} already exists"
            )

        if not entity.product._validator.is_imported(
            entity=entity.product, client=client
        ):
            raise ImportValidationException(
                (
                    f"Product {entity.product.code} is not present in the DB. "
                    "Please import the product first"
                )
            )

        for variable in entity.variables:
            if not variable._validator.is_imported(entity=variable, client=client):
                raise ImportValidationException(
                    (
                        f"Variable {variable.code} is not present in the DB. "
                        "Please import the variable first"
                    )
                )

        if not entity.file_data:
            raise ImportValidationException("File data is missing")

        return True

    def is_imported(self, entity: Union[Experiment, Recipe], client: Client) -> bool:
        """Check if the experiment is already imported"""

        if not entity.id:
            return False

        response = entity.requests(client).get(entity.id)

        if response:
            return True
        return False



""" New file data importers for DHL SDK"""

# pylint: disable=missing-function-docstring, arguments-differ, protected-access
import csv
from io import StringIO
from typing import Any, Dict, Literal, Protocol

from dhl_sdk._constants import FILES_URL
from dhl_sdk.crud import Client


class File(Protocol):
    """Protocol for file data"""

    type: Literal["run", "spectra"]
    variant: Literal["run", "samples"]
    _data: Dict[str, Any]

    def create_request_body(self) -> Dict[str, Any]:
        ...

    def model_dump(
        self,
        exclude_none: bool = ...,
        by_alias: bool = ...,
        include: dict[str, Any] = ...,
    ) -> dict[str, Any]:
        ...


class RunFileImporter:
    """Class to import data from a run file"""

    def __init__(self, client: Client):
        self.client = client

    def import_file(self, file: File) -> str:
        """Import the run file to the project and dataset"""

        response = self.client.post(FILES_URL, file.create_request_body())
        file_id = response.json()["id"]

        import_data = {"timeseries": file._data}

        self.client.put(
            f"{FILES_URL}/{file_id}/data",
            data=import_data,
            content_type="application/json",
        )

        return file_id


class SpectraFileImporter:
    """Class to import data from a spectra file"""

    def __init__(self, client: Client):
        self.client = client

    def import_file(self, file: File) -> tuple[str, str]:
        """Import the spectra file to the project and dataset"""

        # Create Spectra file
        response = self.client.post(FILES_URL, self._file_spectra_body(file))
        spectra_file_id = response.json()["id"]

        import_data = self._spectra_data(file._data, file.variant)

        self.client.put(
            f"{FILES_URL}/{spectra_file_id}/data",
            data=import_data,
            content_type="text/csv",
        )

        # Create Targets file
        response = self.client.post(FILES_URL, self._file_vars_body(file))
        vars_file_id = response.json()["id"]

        import_data = self._variables_data(file._data, file.variant)

        self.client.put(
            f"{FILES_URL}/{vars_file_id}/data",
            data=import_data,
            content_type="application/json",
        )

        return (spectra_file_id, vars_file_id)

    def _file_spectra_body(self, file: File) -> Dict[str, Any]:
        """Create the request body for the spectra file"""
        return file.create_request_body()

    def _file_vars_body(self, file: File) -> Dict[str, Any]:
        """Create the request body for the targets file"""

        file_dict = file.model_dump(
            exclude_none=True,
            by_alias=True,
            include={"name": True, "description": True},
        )

        if file.variant == "run":
            file_dict["type"] = "runData"
        elif file.variant == "samples":
            file_dict["type"] = "sampleData"

        return file_dict

    def _spectra_data(self, data: dict[str, Any], variant: str = "run") -> str:
        """Create the request body for the spectra data"""

        spectra_data = data["spectra"]

        if variant == "run":
            sample_id = "timestamps"
        elif variant == "samples":
            sample_id = "sampleId"
        else:
            raise ValueError(f"Invalid variant: {variant}")

        csv_data = StringIO()
        csv_writer = csv.writer(csv_data)

        # Write data rows
        for timestamp, value_list in zip(
            spectra_data[sample_id], spectra_data["values"]
        ):
            csv_writer.writerow([timestamp] + value_list)

        csv_string = csv_data.getvalue()
        csv_data.close()

        return csv_string

    def _variables_data(self, data: dict[str, Any], variant: str = "run") -> dict:
        """Create the request body for the variables data"""

        data.pop("spectra", None)

        variables_data = data

        if variables_data is None:
            raise ValueError("No spectra data found")

        if variant == "run":
            import_data = {"timeseries": variables_data}
        elif variant == "samples":
            import_data = {"samples": variables_data}
        else:
            raise ValueError(f"Invalid variant: {variant}")

        return import_data



"""
This module defines all the constants used across the project. Constants are 
defined as variables with uppercase names to signify their immutability and 
importance. This file serves as a single source of truth for static values 
to improve code readability, maintainability, and consistency.
"""

PRODUCTS_URL = "api/db/v2/products"
RECIPES_URL = "api/db/v2/recipes"
FILES_URL = "api/db/v2/files"
EXPERIMENTS_URL = "api/db/v2/experiments"
VARIABLES_URL = "api/db/v2/variables"
GROUPS_URL = "api/db/v2/groups"
PROJECTS_URL = "api/db/v2/projects"
DATASETS_URL = "api/db/v2/datasets"
MODELS_URL = "api/db/v2/pipelineJobs"
TEMPLATES_URL = "api/db/v2/pipelineJobTemplates"
PREDICT_URL = "api/pipeline/v1/pipeline"


PROCESS_UNIT_MAP = {
    "cultivation": "04a324da-13a5-470b-94a1-bda6ac87bb86",
    "spectroscopy": "373c173a-1f23-4e56-874e-90ca4702ec0d",
}
